(* 
  talt.v

  A version of TAL with polymorphism (over types), implemented
  using hybrid deBruijn-HOAS method, in which a deBruijn encoding 
  is used for bound variables, and HOAS-style approach is used for
  free variables in open terms.

  Nadeem Abdul Hamid, February 2006
*)

Require Import Compare_dec.
Require Import Eqdep.
Require Import EqNat.
Require Import List.
Require Import natutil.
Require Import tactics.


Open Scope type_scope.

(* TAL_t SYNTAX *)

(* TAL_t registers *)
Inductive reg : Set := r0 | r1 | r2 | r3 | r4 | r5 | r6 | r7.

(* Some utility functions *)
Definition beq_reg_sumbool : forall (r:reg) r', {r=r'}+{r<>r'}.
Proof.
  double induction r r'; intros; auto; try (right; discriminate).
Defined.

Definition beq_reg : forall (r r':reg), bool.
Proof.
  intros; elim (beq_reg_sumbool r r'); intros; [ exact true | exact false ].
Defined.

(* TAL_t TYPES *)

(* Closed types, using deBruijn encoding for bound variables *)
(* The parameter tracks number of free type variables - i.e. the limit
   of valid deBruijn indices *)
Inductive ttypeC : nat -> Set :=
  | tvarC : forall i, ttypeC (S i)
  | tliftC : forall i, ttypeC i -> ttypeC (S i)
  | ttopC : ttypeC 0
  | tintC : ttypeC 0
  | tcodeC : forall i, ctypeC i -> ttypeC i
with ctypeC : nat -> Set :=
  | cbaseC : forall i, rftypeC i -> ctypeC i
  | cbindC  : forall i, ctypeC (S i) -> ctypeC i
with rftypeC : nat -> Set :=
  | rftnilC : forall i, rftypeC i
  | rftconsC : forall i, reg -> ttypeC i -> rftypeC i -> rftypeC i
.


(* Open types, where free TAL_t variables would actually be represented
   using metalevel (Coq) variables *)
Inductive ttype : Set :=
  | ttop : ttype
  | tint : ttype
  | tcode : ctype -> ttype
with ctype : Set :=
  | cbase : rftype -> ctype
  | cbind : ctypeC 1 -> ctype
with rftype : Set :=
  | rftnil : rftype
  | rftcons : reg -> ttype -> rftype -> rftype
.


(* Utility definitions for reasoning about equality in the subst functions *)
Definition myfequal 
  : forall (A B : Type) (f : A -> B) (x y : A), x = y -> f x = f y :=
    fun (A B : Type) (f : A -> B) (x y : A) (H : x = y) =>
      match H in (_ = y0) return (f x = f y0) with
	| refl_equal => refl_equal (f x)
      end.

Definition myeqaddS : forall n m, S n = S m -> n = m.
Proof.
  intros.
  exact (myfequal _ _ (fun e => match e with | 0 => n | S n => n end) _ _ H). 
Defined.

Definition O_S_set : forall (A:Set) j, 0=S j -> A.
Proof.
  intros.
  inversion H. 
Defined.
  

Fixpoint subst_aux (i:nat) (t:ttypeC i) {struct t}
  : forall j, i=(S j) -> ttypeC j -> ttypeC j :=
    match t in (ttypeC i)
      return (forall j, i=S j -> ttypeC j -> ttypeC j) with
      | tvarC n     => fun j _ e => e
      | tliftC n t' => fun j (D:S n=S j) _ => eq_rec n _ t' j (myeqaddS n j D)
      | tintC       => fun j (D:0=S j) _   => O_S_set _ j D
      | ttopC       => fun j (D:0=S j) _   => O_S_set _ j D
      | tcodeC n t' => fun j (D:n=S j) e  => tcodeC j (csubst_aux n t' j D e)
    end
with csubst_aux  (i:nat) (t:ctypeC i) {struct t}
  : forall j, i=(S j) -> ttypeC j -> ctypeC j :=
    match t in (ctypeC i)
      return (forall j, i=S j -> ttypeC j -> ctypeC j) with
      | cbaseC n G => fun j (D:n=S j) e 
	                  => cbaseC j (rfsubst_aux n G j D e)
      | cbindC n t' => fun j (D:n=S j) e
	                  => cbindC j
			          (csubst_aux (S n) t' (S j) 
			                (myfequal _ _ S _ _ D) (tliftC j e))
    end
with rfsubst_aux (i:nat) (G:rftypeC i) {struct G}
  : forall j, i=(S j) -> ttypeC j -> rftypeC j :=
    match G in (rftypeC i)
      return (forall j, i=(S j) -> ttypeC j -> rftypeC j) with
      | rftnilC n => fun j (D:n=S j) _ => rftnilC _
      | rftconsC n r t G' => fun j (D:n=S j) e
	                  => rftconsC j r (subst_aux _ t _ D e)
			                (rfsubst_aux _ G' _ D e)
    end.

Definition tsubst : ttypeC 1 -> ttypeC 0 -> ttypeC 0 :=
  fun t e => (subst_aux _ t _ (refl_equal 1) e).

(*Definition csubst : ctypeC 1 -> ttypeC 0 -> ctypeC 0 :=
  fun t e => (csubst_aux _ t _ (refl_equal 1) e).*)
Parameter csubst : ctypeC 1 -> ttypeC 0 -> ctypeC 0.

Fixpoint unlift_aux i (t:ttypeC i) {struct t} : 0=i -> ttype :=
  match t in (ttypeC i) return (0=i -> ttype) with
    | tvarC n    => fun D => O_S_set _ n D
    | tliftC n _ => fun D => O_S_set _ n D
    | ttopC      => fun _ => ttop
    | tintC      => fun _ => tint
    | tcodeC n t' => fun D => tcode (cunlift_aux n t' D)
  end
with cunlift_aux i (t:ctypeC i) {struct t} : 0=i -> ctype :=
  match t in (ctypeC i) return (0=i -> ctype) with
    | cbaseC n G  => fun D => cbase (rfunlift_aux n G D)
    | cbindC n t' => fun D => cbind (eq_rec (S n) _ t' 1 (sym_eq (eq_S 0 n D)))
  end
with rfunlift_aux i (G:rftypeC i) {struct G} : 0=i -> rftype :=
  match G in (rftypeC i) return (0=i -> rftype) with
    | rftnilC n => fun D => rftnil
    | rftconsC n r t G' => fun D => rftcons r (unlift_aux _ t D)
                                            (rfunlift_aux _ G' D)
  end.

Definition unlift : ttypeC 0 -> ttype := fun t => unlift_aux _ t (refl_equal _).
Definition cunlift : ctypeC 0 -> ctype := 
  fun t => cunlift_aux _ t (refl_equal _).

Fixpoint lift (t:ttype) : ttypeC 0 :=
  match t with
    | ttop => ttopC
    | tint => tintC
    | tcode t' => tcodeC 0 (clift t')
  end
with clift (t:ctype) : ctypeC 0 :=
  match t with
    | cbase G => cbaseC 0 (rflift G)
    | cbind t' => cbindC 0 t'
  end
with rflift (G:rftype) : rftypeC 0 :=
  match G with
    | rftnil => rftnilC _
    | rftcons r t G' => rftconsC _ r (lift t) (rflift G')
  end.

(* register file type *)
Definition rftypeF := reg -> ttype.

(* Code addresses *)
Definition addr := nat.

(* List of types *)
Definition ttypelist := list ttype.

(* word values *)
Inductive wval : Set :=
  | wint : nat -> wval
  | wcaddr : addr -> wval
  | wappt : wval -> ttype -> wval
.

(* register file *)
Definition regfile := reg -> wval.

(* instructions *)
Inductive instr : Set :=
  | iadd : reg -> reg -> reg -> instr
  | iaddi : reg -> reg -> nat -> instr
  | isub  : reg -> reg -> reg -> instr
  | isubi : reg -> reg -> nat -> instr
  | imov  : reg -> reg -> instr
  | imovi : reg -> nat -> instr
  | imovf : reg -> addr -> instr
  | ibgti : reg -> nat -> addr -> ttypelist -> instr
  | itapp : reg -> ttype -> instr
.

(* ibgti and ijd need the natlist to tie things together for the
   preservation lemmas - when doing inversion on evalxm and wf_iseq *)
(* for ijmp we require that the value in the register has
   been fully applied... *)
Inductive iseq : Set :=
  | icons : instr -> iseq -> iseq
  | ijd : addr -> ttypelist -> iseq
  | ijmp : reg -> iseq
.

(* Code instruction sequences under some variable
   environment *)
Inductive ciseq : Set := 
  | cibase : iseq -> ciseq
  | cibind : (ttype -> ciseq) -> ciseq.
(* can't do ciseq (t:ctype) because recursive call not well-founded --
  uses the subst instead of Coq application like TAL_i *)



(* code values are tuples of code type and code instruction sequence *)
Definition codeval := (ctype * ciseq).

(* Code heap *)
Definition codeheap := addr -> option codeval.

(* Program state *)
Definition progstate := (codeheap * regfile * iseq).


(************************************************************************)

(* Register file extensional equality *)
Definition subrftype (G1 G2:rftypeF) : Prop :=
  forall r, (G1 r) = (G2 r) \/ (G2 r) = ttop.

(* Type application definitions *)
Inductive app_ctype : ctype -> ttypelist -> rftype -> Prop :=
  | app_cbase : forall G, app_ctype (cbase G) nil G
  | app_cbind : forall F t ts G,
                app_ctype (cunlift (csubst F (lift t))) ts G ->
		app_ctype (cbind F) (t::ts) G
. (* want to say (F t) in app_cbind *)

Inductive app_ciseq : ciseq -> ttypelist -> iseq -> Prop :=
  | app_cibase : forall Is, app_ciseq (cibase Is) nil Is
  | app_cibind : forall cis t ts Is,
                 app_ciseq (cis t) ts Is ->
		 app_ciseq (cibind cis) (t::ts) Is.



Fixpoint snoc (A:Set) (a:A) (As:list A) {struct As}: list A :=
  match As with
    | nil => a :: nil
    | a' :: As' => a' :: (snoc A a As')
  end.

Fixpoint terase (v:wval) : wval :=
  match v with
    | wappt v _ => terase v
    | v => v
  end.

Fixpoint tcollect (v:wval) : ttypelist :=
  match v with
    | wappt v t => snoc _ t (tcollect v)
    | _ => nil
  end.



(************************************************************************)
(* TAL_i Operational Semantics *)

(* Some utility functions *)

(* Convert rftype to functional representation *)
Fixpoint rftFunc (G:rftype) : reg -> ttype :=
  match G with 
    | rftnil => fun r => ttop
    | rftcons r t G' => fun r' => if (beq_reg r' r) then t 
                                                    else (rftFunc G' r')
  end.


Definition rfupd (R:regfile) (r:reg) (v:wval) : regfile
  := fun r' => if (beq_reg r' r) then v else R r'.


Inductive evalp : progstate -> progstate -> Prop :=
  | ev_iadd : forall C R Is rd rs rt s t,
    R rs = wint s ->
    R rt = wint t ->      
    evalp (C, R, icons (iadd rd rs rt) Is) 
         (C, (rfupd R rd (wint (s+t))), Is)
  | ev_iaddi : forall C R Is rd rs i s,
    R rs = wint s ->
    evalp (C, R, icons (iaddi rd rs i) Is) 
         (C, (rfupd R rd (wint (s+i))), Is)
  | ev_isub : forall C R Is rd rs rt s t,
    R rs = wint s ->
    R rt = wint t ->      
    evalp (C, R, icons (isub rd rs rt) Is) 
         (C, (rfupd R rd (wint (s-t))), Is)
  | ev_isubi : forall C R Is rd rs i s,
    R rs = wint s ->
    evalp (C, R, icons (isubi rd rs i) Is) 
         (C, (rfupd R rd (wint (s-i))), Is)
  | ev_imov : forall C R Is rd rs,
    evalp (C, R, icons (imov rd rs) Is) 
         (C, (rfupd R rd (R rs)), Is)
  | ev_imovi : forall C R Is rd i,
    evalp (C, R, icons (imovi rd i) Is) 
         (C, (rfupd R rd (wint i)), Is)
  | ev_imovf : forall C R Is rd f,
    evalp (C, R, icons (imovf rd f) Is) 
         (C, (rfupd R rd (wcaddr f)), Is)
  | ev_itapp : forall C R Is rd i,
    evalp (C, R, icons (itapp rd i) Is) 
         (C, (rfupd R rd (wappt (R rd) i)), Is)
  | ev_ibgtiT : forall (C:codeheap) R Is rs i f ts s t cis Is',
    R rs = wint s ->
    s > i ->
    C f = Some (t, cis) ->
    app_ciseq cis ts Is' ->
    evalp (C, R, icons (ibgti rs i f ts) Is) (C, R, Is')
  | ev_ibgtiF : forall C R Is rs i f ts s,
    R rs = wint s ->
    s <= i ->
    evalp (C, R, icons (ibgti rs i f ts) Is) (C, R, Is)
  | ev_ijd : forall (C:codeheap) R f ts t cis Is,
    C f = Some (t, cis) ->
    app_ciseq cis ts Is ->
    evalp (C, R, ijd f ts) (C, R, Is)
  | ev_ijmp : forall (C:codeheap) R r f ts t cis Is,
    terase (R r) = wcaddr f ->
    tcollect (R r) = ts ->
    C f = Some (t, cis) ->
    app_ciseq cis ts Is ->
    evalp (C, R, ijmp r) (C, R, Is)
.





(************************************************************************)
(* TAL_i Static Semantics *)

Definition ctypeOf (c:codeval) : ctype := fst c.
(*  match c with cval t _ _ => t end.*)

(* benefit of the approach -- no need of variable environment, or 
   rules for them *)
Inductive wf_wval : codeheap -> wval -> ttype -> Prop :=
  | wf_wtop : forall C v, wf_wval C v ttop
  | wf_wint : forall C i, wf_wval C (wint i) tint
  | wf_wcaddr: forall C f cv,
    C f = Some cv -> wf_wval C (wcaddr f) (tcode (ctypeOf cv))
  | wf_wappt : forall C v t F F',
    wf_wval C v (tcode (cbind F)) ->
    F' = (cunlift (csubst F (lift t))) ->
    wf_wval C (wappt v t) (tcode F')
.

Inductive wf_regfile : codeheap -> regfile -> rftypeF -> Prop :=
  | wf_rfile : forall C R G,
               (forall r, wf_wval C (R r) (G r)) -> wf_regfile C R G
.

Definition rftupd (G:rftypeF) (r:reg) (t:ttype) : rftypeF
  := fun r' => if (beq_reg r' r) then t else G r'.


Inductive wf_iseq : codeheap -> rftypeF -> iseq -> Prop :=
  | wf_iadd : forall C G rd rs rt Is,
    G rs = tint -> G rt = tint ->
    wf_iseq C (rftupd G rd tint) Is ->
    wf_iseq C G (icons (iadd rd rs rt) Is)
  | wf_iaddi : forall C G rd rs t Is,
    G rs = tint -> 
    wf_iseq C (rftupd G rd tint) Is ->
    wf_iseq C G (icons (iaddi rd rs t) Is)
  | wf_isub : forall C G rd rs rt Is,
    G rs = tint -> G rt = tint ->
    wf_iseq C (rftupd G rd tint) Is ->
    wf_iseq C G (icons (isub rd rs rt) Is)
  | wf_isubi : forall C G rd rs t Is,
    G rs = tint -> 
    wf_iseq C (rftupd G rd tint) Is ->
    wf_iseq C G (icons (isubi rd rs t) Is)
  | wf_imov : forall C G rd rs Is,
    wf_iseq C (rftupd G rd (G rs)) Is ->
    wf_iseq C G (icons (imov rd rs) Is)
  | wf_imovi : forall C G rd i Is,
    wf_iseq C (rftupd G rd tint) Is ->
    wf_iseq C G (icons (imovi rd i) Is)
  | wf_imovf : forall C G rd f Is cv,
    C f = Some cv ->
    wf_iseq C (rftupd G rd (tcode (ctypeOf cv))) Is ->
    wf_iseq C G (icons (imovf rd f) Is)
  | wf_itapp : forall C G rd t Is F,
    G rd = tcode (cbind F) ->
    wf_iseq C (rftupd G rd (tcode (cunlift (csubst F (lift t))))) Is ->
    wf_iseq C G (icons (itapp rd t) Is)
  | wf_ibgti : forall C G rs i f ts Is cv G',
    G rs = tint ->
  (* for the true case *)
    C f = Some cv ->
    app_ctype (ctypeOf cv) ts G' ->
    subrftype G (rftFunc G') ->
  (* for the false case *)
    wf_iseq C G Is ->
    wf_iseq C G (icons (ibgti rs i f ts) Is)
  | wf_ijd : forall C G f ts cv G',
    C f = Some cv ->
    app_ctype (ctypeOf cv) ts G' ->
    subrftype G (rftFunc G') ->
    wf_iseq C G (ijd f ts)
  | wf_ijmp : forall C G r G',
    G r = tcode (cbase G') ->
    subrftype G (rftFunc G') ->
    wf_iseq C G (ijmp r)
.


Inductive wf_codeval : codeheap -> codeval -> Prop :=
  wf_cval : forall C t cis,
            (forall ts G, app_ctype t ts G ->
	      exists Is, app_ciseq cis ts Is /\ wf_iseq C (rftFunc G) Is) ->
            wf_codeval C (t, cis).


Inductive wf_codeheap : codeheap -> Prop :=
  | wf_cheap : forall C, 
    (forall f cv, C f = Some cv -> wf_codeval C cv) ->
    wf_codeheap C
.


Inductive wf_progstate : progstate -> Prop :=
  | wf_ps : forall C R Is G,
    wf_codeheap C ->
    wf_regfile C R G ->
    wf_iseq C G Is ->
    wf_progstate (C, R, Is).



(************************************************************************)
(* TAL_i SAFETY PROOFS *)

(* If code type can be instantiated to base, then so can the 
   corresponding code value (used by progress of jump instructions) *)
Lemma app_ctype_ciseq
  : forall C f ct cis ts G,
    wf_codeheap C ->
    C f = Some (ct, cis) ->
    app_ctype ct ts G -> exists Is, app_ciseq cis ts Is.
Proof.
  intros; mcinvClear wf_codeheap.
  generalize (H2 f _ H0); intros HH; inversion HH.
  elim (H4 _ _ H1); introsSplitH; eauto.
Qed.


Lemma app1_snoc : forall A (cs:list A) (c:A), cs ++ c :: nil = snoc _ c cs.
Proof.
  induction cs; unfoldFold snoc; intros; auto.
  replace ((a::cs)++c ::nil) with (a:: (cs++c :: nil)); auto.
  rewrite (IHcs c); auto.
Qed.

Lemma app_nil_eq : forall A (cs:list A), cs++nil = cs.
Proof.
  induction cs; compute; auto.
  unfold app in IHcs; rewrite IHcs; auto.
Qed.

Hint Resolve app1_snoc app_nil_eq.



(* Canoncial forms lemmas *)

Lemma canform_rf_int
  : forall C R G r,
    wf_regfile C R G ->
    G r = tint ->
    exists s, R r = wint s.
Proof.
  intros; mcinvClear wf_regfile.
  generalize (H1 r); mcrewrite (G r); intros; mcinvClear wf_wval.
  eauto.
Qed.


Lemma canform_wf_tcode_aux
  : forall C v t,
    wf_wval C v t ->
    forall ct ts G, t = tcode ct ->
    app_ctype ct ts G ->
    exists f, exists cv, exists ts', exists ts'',
      terase v = wcaddr f /\
      tcollect v = ts' /\
      ts'' = ts' ++ ts /\
      C f = Some cv /\
      app_ctype (ctypeOf cv) ts'' G.
Proof.
  intros C v t H; induction H; intros; try discriminate. 
  existsiv f cv (@nil ttype) ts.
  injection H0; intros; mcrewriteRev ct.
(*  concsplit; auto.
  replace (length ts) with n; auto; apply app_ctype_n_len with ct G; auto.*)

  assert (app_ctype (cbind F) (t::ts) G).
    injection H1; constructor; try mcrewriteRev F'; mcrewriteRev ct.
  generalize (IHwf_wval (cbind F) (t::ts) G (refl_equal _) H3).
    intros (f, (cv, (ts', (ts'', HH)))); splitAllHyps.
  existsiv f cv (ts' ++ t::nil) ts''.
  concsplit; auto.
  unfoldFold tcollect; mcrewrite (tcollect v). symmetry; apply (app1_snoc ttype).
  replace ((ts'++t::nil)++ts) with (ts' ++ (t::nil ++ ts)); auto.
  symmetry; apply (app_ass ts' (t::nil) ts).
Qed.


Lemma canform_wf_tcode
  : forall C v G,
    wf_wval C v (tcode (cbase G)) ->
    exists f, exists cv, exists ts,
      terase v = wcaddr f /\
      tcollect v = ts /\
      C f = Some cv /\
      app_ctype (ctypeOf cv) ts G.
Proof.
  intros.
  generalize
    (canform_wf_tcode_aux C v (tcode (cbase G)) H (cbase G) nil G
      (refl_equal _) (app_cbase _ )).
  intros (f, (cv, (ts', (ts'', HH)))); splitAllHyps.
  existsiii f cv ts''; concsplit; auto.
  mcrewrite ts''. 
  replace (ts'++nil) with ts'; auto.
  symmetry; apply (app_nil_eq ttype).
Qed.


Lemma wf_wval_tcode_app_ctype
  : forall C v G f ts cv,
    wf_wval C v (tcode (cbase G)) ->
    terase v = wcaddr f ->
    tcollect v = ts ->
    C f = Some cv ->
    app_ctype (ctypeOf cv) ts G.
Proof.
  intros.
  generalize (canform_wf_tcode _ _ _ H).
  intros (f', (cv', (ts', HH))); splitAllHyps.
  mcinjection (terase v); mcinjection (tcollect v).
  mcrewriteIn f' (C f'=Some cv'); mcinjection (C f).
  mcrewrite2 cv ts.
Qed.



(* Progress Lemmas *)

Ltac progressStartTac curIs :=
  intros; unfold curIs; mcunfoldIn curIs wf_progstate; clear curIs;
    mcinvClear wf_progstate; mcinvClear wf_iseq.


Lemma progress_iadd
  : forall C R rd rs rt Is,
    let curIs := (icons (iadd rd rs rt) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  elim (canform_rf_int _ _ _ _ H1 H); intro s; intros.
  elim (canform_rf_int _ _ _ _ H1 H3); intro t; intros.
  exists (C, (rfupd R rd (wint (s+t))), Is);
    constructor; auto.
Qed.



Lemma progress_iaddi
  : forall C R rd rs t Is,
    let curIs := (icons (iaddi rd rs t) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  elim (canform_rf_int _ _ _ _ H1 H); intro i; intros.
  exists (C, (rfupd R rd (wint (i+t))), Is);
    constructor; auto.
Qed.


Lemma progress_isub
  : forall C R rd rs rt Is,
    let curIs := (icons (isub rd rs rt) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  elim (canform_rf_int _ _ _ _ H1 H); intro i; intros.
  elim (canform_rf_int _ _ _ _ H1 H3); intro j; intros.
  exists (C, (rfupd R rd (wint (i-j))), Is);
    constructor; auto.
Qed.


Lemma progress_isubi
  : forall C R rd rs t Is,
    let curIs := (icons (isubi rd rs t) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  elim (canform_rf_int _ _ _ _ H1 H); intro i; intros.
  exists (C, (rfupd R rd (wint (i-t))), Is);
    constructor; auto.
Qed.


Lemma progress_imov
  : forall C R rd rs Is,
    let curIs := (icons (imov rd rs) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  exists (C, (rfupd R rd (R rs)), Is);
    constructor; auto.
Qed.


Lemma progress_imovi
  : forall C R rd s Is,
    let curIs := (icons (imovi rd s) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  exists (C, (rfupd R rd (wint s)), Is);
    constructor; auto.
Qed.


Lemma progress_imovf
  : forall C R rd f Is,
    let curIs := (icons (imovf rd f) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  exists (C, (rfupd R rd (wcaddr f)), Is);
    constructor; auto.
Qed.


Lemma progress_itapp
  : forall C R rd t Is,
    let curIs := (icons (itapp rd t) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  exists (C, (rfupd R rd (wappt (R rd) t)), Is);
    constructor; auto.
Qed.


Lemma progress_ibgti
  : forall C R rs i f ts Is,
    let curIs := (icons (ibgti rs i f ts) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  elim (canform_rf_int _ _ _ _ H1 H); intro s; intros.
  
  elim (le_gt_dec s i); intros; auto.
  exists (C, R, Is); eapply ev_ibgtiF; eauto.

  generalize cv H3 H4; clear cv H3 H4; intros (t, cis); unfold ctypeOf, fst; 
    intros.
  elim (app_ctype_ciseq _ _ _ _ _ _ H0 H3 H4); intros Is' His'.
  exists (C, R, Is'); apply ev_ibgtiT with s t cis; auto.
Qed.


Lemma progress_ijd
  : forall C R f ns,
    let curIs := ijd f ns in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  generalize cv H H3; clear cv H H3.
  intros (t, cis); unfold ctypeOf, fst; intros.
  elim (app_ctype_ciseq _ _ _ _ _ _ H0 H H3); unfold projS1; intros Is' His'.
  exists (C, R, Is'); apply ev_ijd with t cis; auto.
Qed.


Lemma progress_ijmp
  : forall C R r,
    let curIs := ijmp r in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  assert (wf_wval C (R r) (tcode (cbase G'))).
    mcinvClear wf_regfile.
    mcrewriteRev (G r); eapplyinf.
  generalize (canform_wf_tcode _ _ _ H2); 
    intros (f, ((t,cis), (ns, HH))); splitAllHyps.
  elim (app_ctype_ciseq _ _ _ _ _ _ H0 H5 H8); 
    unfold ctypeOf, fst; intros Is' His'.
  exists (C, R, Is'); apply ev_ijmp with f ns t cis; auto.
Qed.

  

(* Utility lemmas for preservation proofs *)

Lemma beq_reg_true
  : forall r, true = beq_reg r r.
Proof.
  intros; unfold beq_reg; assert (r=r) ;auto.
  elim (beq_reg_sumbool r r); intros; auto; try absurdEq.
Qed.

Lemma beq_reg_false1
  : forall r r', r<>r' -> false = beq_reg r r'.
Proof.
  intros; unfold beq_reg; elim (beq_reg_sumbool r r'); intros; auto;
    try absurdEq.
Qed.

Lemma beq_reg_false2
  : forall r r', r'<>r -> false = beq_reg r r'.
Proof.
  intros; unfold beq_reg; elim (beq_reg_sumbool r r'); intros; auto;
    try absurdEq.
Qed.

Hint Immediate beq_reg_true.
Hint Resolve beq_reg_false1 beq_reg_false2.

Lemma subrftype_rftupd
  : forall G G' r t,
    subrftype G G' ->
    subrftype (rftupd G r t) (rftupd G' r t).
Proof.
  unfold subrftype; intros G G' r t H r'.
  unfold rftupd; elim (beq_reg r' r); auto.
Qed.

Hint Resolve subrftype_rftupd.

Lemma subrftype_refl
  : forall G, subrftype G G.
Proof.
  intros; intro; auto.
Qed.

Hint Immediate subrftype_refl.

Lemma subrftype_trans
  : forall G G' G'',
    subrftype G G' ->
    subrftype G' G'' ->
    subrftype G G''.
Proof.
  unfold subrftype; intros.
  elim (H r); elim (H0 r); intros;
  elim H1; elim H2; intros; auto.
Qed.

Lemma wf_iseq_subrftype
  : forall C G Is, 
    wf_iseq C G Is -> forall G', subrftype G' G -> wf_iseq C G' Is.
Proof.
  intros C G Is H; induction H; intros;
  repeat match goal with
	   H:G ?r=_, H1:subrftype _ G |- _ => 
	     elim (H1 r); rewrite H; clear H; intros; try discriminate
	 end;
  try constructor; auto.

  apply IHwf_iseq; eauto.
  elim (H0 rs); intros; try mcrewrite (G' rs).
  mcrewrite (G rs); intros; intro r.
    unfold rftupd; elim (beq_reg r rd); auto.

  apply wf_imovf with cv; auto.
  apply wf_itapp with F; auto.
  apply wf_ibgti with cv G'; auto.
    apply subrftype_trans with G; auto.
  apply wf_ijd with cv G'; auto.
    apply subrftype_trans with G; auto.
  apply wf_ijmp with G'; auto.
    apply subrftype_trans with G; auto.
Qed.

Lemma app_ciseq_func
  : forall cis ts Is,
    app_ciseq cis ts Is ->
    forall Is',
    app_ciseq cis ts Is' ->
    Is = Is'.
Proof.
  intros cis ts Is H; induction H; intros; mcinv app_ciseq.
Qed.




(* Preservation Lemmas *)

Definition preservStatement C R i Is P'
  := let curIs := icons i Is in
     wf_progstate (C, R, curIs) -> evalp (C, R, curIs) P' ->
     wf_progstate P'.

Lemma preserv_iadd
  : forall C R rd rs rt Is P',
    preservStatement C R (iadd rd rs rt) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd tint); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
Qed.


Lemma preserv_iaddi
  : forall C R rd rs i Is P',
    preservStatement C R (iaddi rd rs i) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd tint); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
Qed.


Lemma preserv_isub
  : forall C R rd rs rt Is P',
    preservStatement C R (isub rd rs rt) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd tint); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
Qed.


Lemma preserv_isubi
  : forall C R rd rs i Is P',
    preservStatement C R (isubi rd rs i) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd tint); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
Qed.


Lemma preserv_imov
  : forall C R rd rs Is P',
    preservStatement C R (imov rd rs) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd (G rs)); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; mcinvClear wf_regfile.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
Qed.


Lemma preserv_imovi
  : forall C R rd i Is P',
    preservStatement C R (imovi rd i) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd tint); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
Qed.


Lemma preserv_imovf
  : forall C R rd f Is P',
    preservStatement C R (imovf rd f) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  mcinvClear wf_iseq.
  apply wf_ps with (rftupd G rd (tcode (ctypeOf cv))); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor; auto.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
Qed.


Lemma preserv_itapp
  : forall C R rd n Is P',
    preservStatement C R (itapp rd n) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  mcinvClear wf_iseq.
  apply wf_ps with (rftupd G rd (tcode (cunlift (csubst F (lift n)))));
    intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto. 
  apply wf_wappt with F; auto.
    mcrewriteRev (G rd); mcinvClear wf_regfile.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
Qed.


Lemma preserv_ibgti
  : forall C R rs i f ts Is P',
    preservStatement C R (ibgti rs i f ts) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros.
  assert (C'=C).
    mcinv evalp.
  mcrewrite C'; mcrewriteIn C' evalp; mcclear C'.

  mcinv evalp;

  mcclear5 C0 R0 rs0 i0 f0; mcclear3 ts0 Is0 Is'0;
  mcinvClear wf_progstate; mcinvClear wf_iseq;
  apply wf_ps with G; auto;
  try mcrewrite R'; try mcrewrite Is'.
  
  assert (wf_codeval C cv).
     mcinv wf_codeheap; eapplyinf.
  mcinjection (C f).
  mcrewriteIn cv wf_codeval; mcinvClear wf_codeval.
  mcrewriteIn cv app_ctype; mcunfoldIn ctypeOf app_ctype.
  elim (H14 _ _ H6); introsSplitH.

  apply wf_iseq_subrftype with (rftFunc G'); auto.
  replace Is' with x; auto.
 
  apply app_ciseq_func with cis ts; auto.
Qed.



Lemma preserv_ijd
  : forall C R f ts P',
    let curIs := (ijd f ts) in
      wf_progstate (C, R, curIs) -> evalp (C, R, curIs) P' ->
      wf_progstate P'.
Proof.
  intros C R f ts ((C', R'), Is'); intros.
  assert (C'=C).
    mcinv evalp.
  mcrewrite C'; mcrewriteIn C' evalp; mcclear C'.

  mcinv evalp;

  mcclear5 C0 R0 rs0 i0 f0; mcclear3 ts0 Is0 Is'0;
  mcinvClear wf_progstate; mcinvClear wf_iseq;
  apply wf_ps with G; auto;
  try mcrewrite R'; try mcrewrite Is'.
  
  assert (wf_codeval C cv).
     mcinv wf_codeheap; eapplyinf.
  mcinjection (C f).
  mcrewriteIn cv wf_codeval; mcinvClear wf_codeval.
  mcrewriteIn cv app_ctype; mcunfoldIn ctypeOf app_ctype.
  elim (H11 _ _ H5); introsSplitH.

  apply wf_iseq_subrftype with (rftFunc G'); auto.
  replace Is with Is';
  replace Is' with x; auto.
 
  apply app_ciseq_func with cis ts; auto.
Qed.


Lemma preserv_ijmp
  : forall C R r P',
    let curIs := (ijmp r) in
      wf_progstate (C, R, curIs) -> evalp (C, R, curIs) P' ->
      wf_progstate P'.
Proof.
  intros C R r ((C', R'), Is'); intros.
  assert (C'=C).
    mcinv evalp.
  mcrewrite C'; mcrewriteIn C' evalp; mcclear C'.

  mcinv evalp;

  mcclear5 C0 R0 r8 i0 f0; mcclear3 ns0 Is0 Is'0;
  mcinvClear wf_progstate; mcinvClear wf_iseq;
  apply wf_ps with G; auto;
  try mcrewrite R'.
  
  mcinvClear wf_codeheap. generalize (H3 _ _ H8); intros.
  mcinvClear wf_codeval.
  
  assert (app_ctype t ts G'). 
    mcinvClear wf_regfile.
    generalize (H1 r); mcrewrite (G r); intros.
    repeat mcrewriteIn R' eq; mcrewriteIn R' evalp; mcclear R'.
    generalize (wf_wval_tcode_app_ctype _ _ _ _ _ _ H2 H6 H7 H8);
      unfold ctypeOf, fst; intros; auto.
  elim (H11 _ _ H1); introsSplitH.
  apply wf_iseq_subrftype with (rftFunc G'); auto.
  replace Is' with x; auto.
  apply app_ciseq_func with cis ts; auto.
Qed.


(* Complete theorems *)

Theorem progress 
  : forall P, wf_progstate P -> exists P', evalp P P'.
Proof.
  intros ((C, R), Is); elim Is; clear Is. 

  intro i; case i; intros.
  eapply progress_iadd; eauto.
  eapply progress_iaddi; eauto.
  eapply progress_isub; eauto.
  eapply progress_isubi; eauto.
  eapply progress_imov; eauto.
  eapply progress_imovi; eauto.
  eapply progress_imovf; eauto.
  eapply progress_ibgti; eauto.
  eapply progress_itapp; eauto.
  intros; eapply progress_ijd; eauto.
  intros; eapply progress_ijmp; eauto.
Qed.

Theorem preservation 
  : forall P P', wf_progstate P -> evalp P P' -> wf_progstate P'.
Proof.
  intros ((C, R), Is); elim Is; clear Is.

  intro i; case i; intros.
  generalize preserv_iadd; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_iaddi; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_isub; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_isubi; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_imov; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_imovi; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_imovf; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_ibgti; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_itapp; unfold preservStatement; intros HH; eapply HH; eauto.
  intros; eapply preserv_ijd; eauto.
  intros; eapply preserv_ijmp; eauto.
Qed.









(*******)

Scheme tInd := Induction for ttype Sort Prop
with ctInd := Induction for ctype Sort Prop
with rftInd := Induction for rftype Sort Prop.


Lemma unlift_lift_eq
  : forall t, unlift_aux _ (lift t) (refl_equal _) = t.
Proof.
  intros t.
  elim t using tInd
    with (P:=(fun t=>unlift_aux _ (lift t) (refl_equal _) = t))
         (P0:=(fun t=>cunlift_aux _ (clift t) (refl_equal _) = t))
         (P1:=(fun t=>rfunlift_aux _ (rflift t) (refl_equal _) = t));
	 intros; auto.
  unfoldFold lift.
  replace (tcode c) with (tcode (cunlift_aux 0 (clift c) (refl_equal 0))).
  compute; auto.
  rewrite H; auto.

  unfoldFold clift.
  replace (cbase r) with (cbase ( rfunlift_aux 0 (rflift r) (refl_equal 0))).
  compute; auto.
  rewrite H; auto.

  unfoldFold rflift.
  replace (rftcons r t0 r8) with
    (rftcons r ( unlift_aux 0 (lift t0) (refl_equal 0)) 
      ( rfunlift_aux 0 (rflift r8) (refl_equal 0))).
  compute; auto.
  rewrite H; rewrite H0; auto.
Qed.


Lemma cunlift_lift_eq
  : forall t, cunlift_aux _ (clift t) (refl_equal _) = t.
Proof.
  intros t.
  elim t using ctInd
    with (P:=(fun t=>unlift_aux _ (lift t) (refl_equal _) = t))
         (P0:=(fun t=>cunlift_aux _ (clift t) (refl_equal _) = t))
         (P1:=(fun t=>rfunlift_aux _ (rflift t) (refl_equal _) = t));
	 intros; auto.
  unfoldFold lift.
  replace (tcode c) with (tcode (cunlift_aux 0 (clift c) (refl_equal 0))).
  compute; auto.
  rewrite H; auto.

  unfoldFold clift.
  replace (cbase r) with (cbase ( rfunlift_aux 0 (rflift r) (refl_equal 0))).
  compute; auto.
  rewrite H; auto.

  unfoldFold rflift.
  replace (rftcons r t0 r8) with
    (rftcons r ( unlift_aux 0 (lift t0) (refl_equal 0)) 
      ( rfunlift_aux 0 (rflift r8) (refl_equal 0))).
  compute; auto.
  rewrite H; rewrite H0; auto.
Qed.


Lemma rfunlift_lift_eq
  : forall t, rfunlift_aux _ (rflift t) (refl_equal _) = t.
Proof.
  intros t.
  elim t using rftInd
    with (P:=(fun t=>unlift_aux _ (lift t) (refl_equal _) = t))
         (P0:=(fun t=>cunlift_aux _ (clift t) (refl_equal _) = t))
         (P1:=(fun t=>rfunlift_aux _ (rflift t) (refl_equal _) = t));
	 intros; auto.
  unfoldFold lift.
  replace (tcode c) with (tcode (cunlift_aux 0 (clift c) (refl_equal 0))).
  compute; auto.
  rewrite H; auto.

  unfoldFold clift.
  replace (cbase r) with (cbase ( rfunlift_aux 0 (rflift r) (refl_equal 0))).
  compute; auto.
  rewrite H; auto.

  unfoldFold rflift.
  replace (rftcons r t0 r8) with
    (rftcons r ( unlift_aux 0 (lift t0) (refl_equal 0)) 
      ( rfunlift_aux 0 (rflift r8) (refl_equal 0))).
  compute; auto.
  rewrite H; rewrite H0; auto.
Qed.

